#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
import os
import sys
import time
import functools
import threading
import string

import serial

import indigo

### NOTES on logging:
###		LOTS of debugLog commands are placed throughout.  These can be activated in plugin.py
###			to "debug" misbahavior between Indigo <-> Aqualink Serial <-> Jandy System.
###		The device state "loggingChoice" is used to determine logging for "typical" communciations
###		A very few direct indigo.server.log commands are included where logging seemed essential.

##################################################################################################
# 1 sec seems like a balanced sleep after commands between
# the USB-to-Serial and Aqualink Serial RS connection.  It's slooooow.
kSleepAfterSerialCommand = 1.00


##################################################################################################
class JandyAqualink(object):

	##############################################################################################
	def __init__(self, plugin):
		self.plugin = plugin
	
	########################################
	def __del__(self):
		pass
			
	##############################################################################################
	def startCommThread(self, dev):
		
		devProps = dev.pluginProps
		portName = devProps.get("serialPort")
		self.plugin.debugLog("Serial Port Name is " + portName)
		self.conn = self.plugin.openSerial(dev.name, portName, 9600, timeout=1, writeTimeout=1)

		### If connection to serial device is successfully statrted, queue initial commands
		### Else, log error starting communication
		if self.conn:
			self.plugin.debugLog("Device name is " + dev.name)

###!!!	NOTE:	Still need to set up POOLHT2, POOLSP2; states not yet established.
###!!!			POOLHT2 only be avalable in Pool Only / Spa Only systems, which are not yet configured.

			###	Set value for Jandy Logging based on Device Configuration
			logChoice = devProps.get("logChoiceMenu")
			if logChoice == "logOff":
				dev.updateStateOnServer("loggingChoice", value="logOff")
				indigo.server.log("Logging of Jandy state changes turned off.")
			elif logChoice == "logAllButTemp":
				dev.updateStateOnServer("loggingChoice", value="logAllButTemp")
				indigo.server.log("Logging of Jandy state changes set to log all but temperature changes.")
			else:		# Default to log all - even if no selection was made in set-up dialog.
				dev.updateStateOnServer("loggingChoice", value="logAll")
				indigo.server.log("Logging of Jandy state changes set to log all changes.")

			### Queue commands to initialize Aqualink serial communications and initialize states.
			####	UNITS=0		Get temperature units and store so that it can be added to state info.
			###		COSMSGS=1	Set continuous messaging from Jandy; easiest way to keep temperatures current.
			###					uses the first portion of the concurrent thread to read / clean up.
			###	Set value for Jandy Logging to log all during initial communications.
			self.commQueue = []
			self.queueSerialCmd("#MODEL?")		# May use Model and Options in the future to
#			self.queueSerialCmd("#OPTIONS?")	# validate or set up device w/o user intervention.
			self.queueSerialCmd("#UNITS?")
			self.queueSerialCmd("#LEDS?")
			if dev.deviceTypeId[19:23] == "Pool":
				if devProps["systemChoice"] == "pool":			
					self.queueSerialCmd("#POOLSP?")
					self.queueSerialCmd("#POOLTMP?")
				else:
					self.queueSerialCmd("#SPASP?")
					self.queueSerialCmd("#SPATMP?")
			else:
				self.queueSerialCmd("#POOLSP?")
				self.queueSerialCmd("#SPASP?")
				self.queueSerialCmd("#POOLTMP?")
				self.queueSerialCmd("#SPATMP?")
			if devProps["solHeatCheckbox"] == True:			
				self.queueSerialCmd("#SOLTMP?")
			self.queueSerialCmd("#AIRTMP?")
			if devProps["saltSystem"] == True:			
				self.queueSerialCmd("#SALT?")
			self.queueSerialCmd("#COSMSGS=1")			
			for command in self.commQueue:
				self.plugin.debugLog("In queue: " + command)				

			### Start separate concurrentSerialComm thread - syntax roughly based on EasyDAQ Plugin
			devId = dev.id
			self.concurThread = threading.Thread(target=functools.partial(self.concurrentSerialComm, devId))
			self.concurThread.start()
			self.plugin.debugLog("Started concurrent thread.")
			
		else:
			indigo.server.log("Error initializing communciations with serial device " + dev.name)
			
	######################
	# Send command to stop concurrentSerialComm thread, then close serial connection and exit.
	def stopCommThread(self, dev):
		self.plugin.debugLog("Initiating stop of concurrent thread.")
		self.queueSerialCmd("stopConcurrentSerialComm")
		time.sleep(10)		# Give plenty of time for command to stop concurrent thread to execute.
		self.conn.close()
		self.plugin.debugLog("closed connection to device " + dev.name)


	##############################################################################################
	def concurrentSerialComm(self, devId):
		try:
			
			self.plugin.debugLog("Starting concurrent serial communications.")
			while True:

				##################################################################################
				# First on each pass, get current device state infomation, as it may have changed,
				# 	and needs to be used to ensure proper commands for some state updates.
				dev = indigo.devices[devId]

				##################################################################################
				# Next, on each pass read/decode any commands from Aqualink.
				# Shouldn't have anything until/unless COSMSGS=1 is sent to start ongoing communication.
				self.readSerialBuffer(devId)
#				self.plugin.debugLog("Reading serial buffer complete.")

				##################################################################################
				# Next, send commands from those queued.
				# Use a copy of the queue to avoid confusion if added to while executing.
				# Check for command to terminate thread
				# Clear commands executed here, but not any added while executing.
				# Only allow temperature requsts if pump is on (all systems) and valves are in proper position (Combo systems).
				# Note that user could send other commands that are not allowed, 
				#	but will receive Jandy error message in response, i.e. not all possibilities are screened.
				if len(self.commQueue) >= 1:
					self.plugin.debugLog("Queue has " + str(len(self.commQueue)) + " command(s) waiting.")
					workingQueue = self.commQueue
					if "stopConcurrentSerialComm" in workingQueue:	# Command to stop queue?
						self.plugin.debugLog("Raising exception to stop concurrent thread.")
						workingQueue = []
						self.commQueue = []			#Clear queue
						raise self.StopThread		# Raise exception to stop thread.
					lenQueue = len(workingQueue)
					for command in workingQueue:
						self.plugin.debugLog("Processing command: " + command)
						if command == "#POOLTMP?":
							if dev.deviceTypeId == "jandyAqualink-RS-2-6-SepEquip":
								currentPumpState = dev.states["poolPumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("poolTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Pool pump off, cannot determine pool temperature.")
								else:
									self.sendSerialCmd(devId, command)
									self.plugin.debugLog(command + " command completed.")
							elif dev.deviceTypeId[19:24] == "Combo":
								currentPumpState = dev.states["pumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("poolTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Pump off or switching, cannot determine pool temperature.")
								else:
									currentValveState = dev.states["valveState"]
									if currentValveState == "pool":
										self.sendSerialCmd(devId, command)
										self.plugin.debugLog(command + " command completed.")
									else:
										dev.updateStateOnServer("poolTemp", value="--")
										if dev.states["loggingChoice"] != "logOff":
											indigo.server.log("Valves set to " + currentValveState + ".  Cannot determine pool temperature; set to --.")
							else:
								currentPumpState = dev.states["pumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("poolTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Pump off, cannot determine pool temperature.")
								else:
									self.sendSerialCmd(devId, command)
									self.plugin.debugLog(command + " command completed.")
						elif command == "#SPATMP?":
							if dev.deviceTypeId == "jandyAqualink-RS-2-6-SepEquip":
								currentPumpState = dev.states["spaPumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("spaTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Spa pump off, cannot determine spa temperature.")
								else:
									self.sendSerialCmd(devId, command)
									self.plugin.debugLog(command + " command completed.")
							elif dev.deviceTypeId[19:24] == "Combo":
								currentPumpState = dev.states["pumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("spaTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Pump off or switching, cannot determine spa temperature.")
								else:
									currentValveState = dev.states["valveState"]
									if currentValveState == "spa":
										self.sendSerialCmd(devId, command)
										self.plugin.debugLog(command + " command completed.")
									else:
										dev.updateStateOnServer("spaTemp", value="--")
										if dev.states["loggingChoice"] != "logOff":
											indigo.server.log("Valves set to " + currentValveState + ".  Cannot determine spa temperature; set to --.")
							else:
								currentPumpState = dev.states["pumpState"]
								if currentPumpState == "off" or currentPumpState == "waitingForDelay":
									dev.updateStateOnServer("spaTemp", value="--")
									if dev.states["loggingChoice"] != "logOff":
										indigo.server.log("Pump off, cannot determine spa temperature.")
								else:
									self.sendSerialCmd(devId, command)
									self.plugin.debugLog(command + " command completed.")
						else:
							self.sendSerialCmd(devId, command)
							self.plugin.debugLog(command + " command completed.")
					self.commQueue = self.commQueue[lenQueue:]

				##################################################################################
				# Sleep a bit on each cycle.
				time.sleep(kSleepAfterSerialCommand * 3)
#				self.plugin.debugLog("Routine cycle through concurrentSerialComm thread.")
				
		### Exceptions to exit thread.  !!! Need to understand 2nd and 3rd except:, "borrowed" from EasyDAQ plugin.
		except self.StopThread:
			pass	# silently fall into finally: section below
		except Exception, e:
			self.plugin.exceptionLog()
		except:
			self.plugin.exceptionLog()
		finally:
			self.plugin.debugLog("Should now stop concurrent thread.")
			pass	# Finally, exit thread.
			

	##############################################################################################
	### Add new command to queue, which is polled and emptied by concurrentSerialComm funtion
	def queueSerialCmd(self, command):
		self.commQueue.append(command)	# Append new command.
		

	##############################################################################################
	### Check to see if Jandy system has sent any information; will only occur if COSMSGS set on.
	def readSerialBuffer(self, devId):
		dev = indigo.devices[devId]
		dev.updateStateOnServer("commActive", value="true")
		charNum = self.conn.inWaiting()
#		self.plugin.debugLog("Number of characters waiting is " + str(charNum))
		time.sleep(kSleepAfterSerialCommand)
		dev.updateStateOnServer("commActive", value="false")
		if charNum > 0:
			dev.updateStateOnServer("commActive", value="true")
			responseString = self.conn.read(size=charNum)
			time.sleep(kSleepAfterSerialCommand)
			dev.updateStateOnServer("commActive", value="false")
			self.plugin.debugLog("Jandy sent string " + responseString)
			commDirection = "Read"
			self.decodeResponse(devId, responseString, commDirection)

	########################################
	### Send queued command to Jandy system, then read reply.
	def sendSerialCmd(self, devId, command):
		dev = indigo.devices[devId]
		writeCommand = command + "\r"	# append carraige return
		dev.updateStateOnServer("commActive", value="true")
		sentCount = self.conn.write(writeCommand)
		time.sleep(kSleepAfterSerialCommand)
		charNum = self.conn.inWaiting()
#		self.plugin.debugLog("Number of characters waiting is " + str(charNum))
		time.sleep(kSleepAfterSerialCommand)
		responseString = self.conn.read(size=charNum)
		time.sleep(kSleepAfterSerialCommand)
		dev.updateStateOnServer("commActive", value="false")
#		self.plugin.debugLog("Returned string length " + str(len(responseString)))
		self.plugin.debugLog("Returned string is " + responseString)
		commDirection = "Send"
		self.decodeResponse(devId, responseString, commDirection)
		

	##############################################################################################
	### Decode response
	### May have more than one item in response string, even if sending command, especially if COSMSGS = 1
	### Anything sent from Jandy indiates state change, so change state variable
	###	Actions are also handled to only be sent to Jandy if resulting in state change
	### Log state changes based on "loggingChoice" state
	def decodeResponse(self, devId, responseString, commDirection):
			
			stringLen = len(responseString)
			while stringLen > 0:

				self.plugin.debugLog("Remaining response to decode: " + responseString + ", of length " + str(stringLen) + " characters")
				### Get fresh version of device states, as some are decoded based on others.			
				dev = indigo.devices[devId]
				

				### All responses should end with <CR> <LF>, but occasionally one comes through that's truncated?
				### If no <LF> found at the end, then response is faulty, abort this decode
				if string.find(responseString,u"\u000A") == -1:
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Faulty response: " + responseString)
					stringLen = 0
					responseString = []
					
				### If LED command was sent, decode separately		
				elif responseString[:10] == "!00 LEDS =":
					self.decodeLED(devId, responseString)
					lineFeed = string.find(responseString,u"\u000A") + 1
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]

				### Pump changes
				### The only Pump for Pool/Spa Combo and Pool or Spa Only systems, the Pool Pump for Pool/Spa Separate systems
				### After making initial update, queue LED command, as COSMSGS does not return full changes while valve switching.
				###	PUMPLO command/response will only function if this option is installed on AUX2.
				elif responseString[:10] == "!00 PUMP =":
					newState = responseString[11]
					if newState == "0":
						newState = "off"
					else:
						newState = "on"
					if dev.deviceTypeId != "jandyAqualink-RS-2-6-SepEquip":
						currentState = dev.states["pumpState"]
						dev.updateStateOnServer("pumpState", value=newState)
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("Pump state: " + newState)
						if newState != currentState:
							self.queueSerialCmd("#LEDS?")
						stringLen = stringLen - 14
						self.plugin.debugLog(str(stringLen))
						if stringLen > 0:
							responseString = responseString[14:]		
					else:
						currentState = dev.states["poolPumpState"]
						dev.updateStateOnServer("poolPumpState", value=newState)
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("Pool pump state: " + newState)
						if newState != currentState:
							self.queueSerialCmd("#LEDS?")
						stringLen = stringLen - 14
						self.plugin.debugLog(str(stringLen))
						if stringLen > 0:
							responseString = responseString[14:]		
####!!!! NOTE: This command/response has not been properly tested.
				elif responseString[:12] == "!00 PUMPLO =":
					newState = responseString[13]
					if newState == "0":
						newState = "off"
					else:
						newState = "on"
					dev.updateStateOnServer("pumpLoState", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Low speed pump state: " + newState)
					stringLen = stringLen - 16
					if stringLen > 0:
						responseString = responseString[16:]

				### Valve changes (valveState) for Combo systems
				### or Spa pump state (spaPumpState) for RS 2/6 separate pool/spa system
				### This command should not occur for Pool or Spa Only systems, which have only pump on/off.	
				### Note that valve states are referenced to spa, on or off.
				### After making initial update, queue LED command, as COSMSGS does not return full changes while valve switching.
				### Also set Temperature (for COMBO systems) to "--" for inactive pool or spa; active area will update with COSMSGS
				elif responseString[:9] == "!00 SPA =":
					newState = responseString[10]
					if dev.deviceTypeId[19:24] == "Combo":
						currentState = dev.states["valveState"]
						if newState == "0":
							newState = "pool"
							dev.updateStateOnServer("spaTemp", value="--")
						else:
							newState = "spa"
							dev.updateStateOnServer("poolTemp", value="--")
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("Valves switched/switching to " + newState)
						if newState != currentState:
							self.queueSerialCmd("#LEDS?")
						stringLen = stringLen - 13
						if stringLen > 0:
							responseString = responseString[13:]
					else:
						currentState = dev.states["spaPumpState"]
						if newState == "0":
							newState = "off"
							dev.updateStateOnServer("spaTemp", value="--")
						else:
							newState = "on"
						dev.updateStateOnServer("spaPumpState", value=newState)
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("Spa Pump switched " + newState)
						if newState != currentState:
							self.queueSerialCmd("#LEDS?")
						stringLen = stringLen - 13
						if stringLen > 0:
							responseString = responseString[13:]					

				### Heater changes
				### For Combo systems: Returned value can be 0=off or 1=enabled or 2=on
				### If "on," valveState determines whether on or enabled.
				### For Separate Equipment (RS 2/6) and Pool or Spa Only: Returned state is on/off
				###	Note that POOLHT2 command included (pool only / spa only option), 
				###		but will not function unless installed.  It will be returned in temperature units are changed.
				###	Note that SOLHT command/response will only function if this option is properly installed.
				elif responseString[:12] == "!00 POOLHT =":
					heaterState = responseString[13]
					if dev.deviceTypeId[19:24] == "Combo":
						valveState = dev.states["valveState"]
						if heaterState == "0":
							newState = "off"
						elif heaterState == "1" or valveState == "spa":
							newState = "enabled"
						elif heaterState == "2":
							newState = "on"
					else:
						if heaterState == "0":
							newState = "off"
						elif heaterState == "1":
							newState = "enabled"
						else:
							newState = "on"
					dev.updateStateOnServer("poolHeat", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Pool heat: " + newState)
					stringLen = stringLen - 16
					if stringLen > 0:
						responseString = responseString[16:]
####!!!! NOTE: This command/response has not been properly tested.
				elif responseString[:13] == "!00 POOLHT2 =":
					heaterState = responseString[14]
					if dev.deviceTypeId[19:24] == "Combo":
						valveState = dev.states["valveState"]
						if heaterState == "0":
							newState = "off"
						elif heaterState == "1" or valveState == "spa":
							newState = "enabled"
						elif heaterState == "2":
							newState = "on"
					else:
						if heaterState == "0":
							newState = "off"
						elif heaterState == "1":
							newState = "enabled"
						else:
							newState = "on"
					dev.updateStateOnServer("poolHeat2", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Pool heat 2: " + newState)
					stringLen = stringLen - 17
					if stringLen > 0:
						responseString = responseString[17:]
				elif responseString[:11] == "!00 SPAHT =":
					heaterState = responseString[12]
					if dev.deviceTypeId[19:24] == "Combo":
						valveState = dev.states["valveState"]
						if heaterState == "0":
							newState = "off"
						elif heaterState == "1" or valveState == "pool":
							newState = "enabled"
						elif heaterState == "2":
							newState = "on"
					else:
						if heaterState == "0":
							newState = "off"
						else:
							newState = "on"
					dev.updateStateOnServer("spaHeat", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Spa heat: " + newState)
					stringLen = stringLen - 15
					if stringLen > 0:
						responseString = responseString[15:]
####!!!! NOTE: This command/response has not been properly tested.
				elif responseString[:11] == "!00 SOLHT =":
					heaterState = responseString[12]
					if heaterState == "0":
						newState = "off"
					elif heaterState == "1":
						newState = "active"
					elif heaterState == "2":
						newState = "enabled"
					dev.updateStateOnServer("solHeat", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Solar heat: " + newState)
					stringLen = stringLen - 15
					if stringLen > 0:
						responseString = responseString[15:]

				### Temperature changes		
				###	Note that SOLTMP command/response will only function if this option is properly installed.
				elif responseString[:13] == "!00 POOLTMP =":
					extraDigit = 0
					if responseString[16] != " ":
						extraDigit = 1
					newTemp = int(responseString[14:16+extraDigit])
					dev.updateStateOnServer("poolTemp", value=newTemp)
					if dev.states["loggingChoice"] == "logAll":
						indigo.server.log("Pool temperature: " + str(newTemp))
					stringLen = stringLen - (20 + extraDigit)
					if stringLen > 0:
						responseString = responseString[20+extraDigit:]
				elif responseString[:12] == "!00 SPATMP =":
					extraDigit = 0
					if responseString[15] != " ":
						extraDigit = 1
					newTemp = int(responseString[13:15+extraDigit])
					dev.updateStateOnServer("spaTemp", value=newTemp)
					if dev.states["loggingChoice"] == "logAll":
						indigo.server.log("Spa temperature: " + str(newTemp))
					stringLen = stringLen - (19 + extraDigit)
					if stringLen > 0:
						responseString = responseString[19+extraDigit:]
####!!!! NOTE: This command/response has not been properly tested.
				elif responseString[:12] == "!00 SOLTMP =":
					extraDigit = 0
					if responseString[15] != " ":
						extraDigit = 1
					newTemp = int(responseString[13:15+extraDigit])
					dev.updateStateOnServer("solTemp", value=newTemp)
					if dev.states["loggingChoice"] == "logAll":
						indigo.server.log("Solar temperature: " + str(newTemp))
					stringLen = stringLen - (19 + extraDigit)
					if stringLen > 0:
						responseString = responseString[19+extraDigit:]
				elif responseString[:12] == "!00 AIRTMP =":
					extraDigit = 0
					if responseString[15] != " ":
						extraDigit = 1
					newTemp = int(responseString[13:15+extraDigit])
					dev.updateStateOnServer("airTemp", value=newTemp)
					if dev.states["loggingChoice"] == "logAll":
						indigo.server.log("Air temperature: " + str(newTemp))
					stringLen = stringLen - (19 + extraDigit)
					if stringLen > 0:
						responseString = responseString[19+extraDigit:]

				### Temperature setpoint changes		
				###	Note that POOLSP2 command included (pool only / spa only option), 
				###		but will not function unless properly installed.
				elif responseString[:12] == "!00 POOLSP =":
					extraDigit = 0
					if responseString[15] != " ":
						extraDigit = 1
					newTemp = int(responseString[13:15+extraDigit])
					dev.updateStateOnServer("poolSetpoint", value=newTemp)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Pool setpoint: " + str(newTemp))
					stringLen = stringLen - (19 + extraDigit)
					if stringLen > 0:
						responseString = responseString[19+extraDigit:]
####!!!! NOTE: This command/response has not been properly tested.
				### Note that POOLSP2 willbe returned for any temperature unit change, 
				###		but state is only used for pool/spa only systems.
				elif responseString[:13] == "!00 POOLSP2 =":
					extraDigit = 0
					if responseString[16] != " ":
						extraDigit = 1
					if "poolSetpoint2" in dev.states:
						newTemp = int(responseString[14:16+extraDigit])
						dev.updateStateOnServer("poolSetpoint2", value=newTemp)
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("Pool setpoint 2: " + str(newTemp))
					stringLen = stringLen - (20 + extraDigit)
					if stringLen > 0:
						responseString = responseString[20+extraDigit:]
				elif responseString[:11] == "!00 SPASP =":
					extraDigit = 0
					if responseString[14] != " ":
						extraDigit = 1
					newTemp = int(responseString[12:14+extraDigit])
					dev.updateStateOnServer("spaSetpoint", value=newTemp)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Spa setpoint: " + str(newTemp))
					stringLen = stringLen - (18 + extraDigit)
					if stringLen > 0:
						responseString = responseString[18+extraDigit:]

				### CLEANR changes; will only function if pool cleaner is properly installed on AUX1
				elif responseString[:12] == "!00 CLEANR =":
					newState = responseString[13]
					if newState == "0":
						newState = "off"
					else:
						newState = "on"
					dev.updateStateOnServer("cleanr", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Cleaner: " + newState)
					stringLen = stringLen - 16
					if stringLen > 0:
						responseString = responseString[16:]

				### WFALL changes; will only function if waterfall feature is properly installed on AUX3
				### If waterfall feature is installed on another auxilliary, 
				###		it will function with a standard AUX command
####!!!! NOTE: This command/response has not been properly tested.
				elif responseString[:11] == "!00 WFALL =":
					newState = responseString[12]
					if newState == "0":
						newState = "off"
					else:
						newState = "on"
					dev.updateStateOnServer("waterfall", value=newState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Waterfall: " + newState)
					stringLen = stringLen - 15
					if stringLen > 0:
						responseString = responseString[15:]

				### Auxillary changes
				elif responseString[:7] == "!00 AUX":
					auxID = responseString[7]
					if responseString[12] != " ":
						auxState = responseString[11]
						if auxState == "0":
							auxState = "off"
						else:
							auxState = "on"
						stringLen = stringLen - 14
						if stringLen > 0:
							responseString = responseString[14:]
####!!!! NOTE: Decoding for dimming responses have not been validated.
					else:
						lineFeed = string.find(responseString,u"\u000A") + 1
						auxState = responseString[13:lineFeed]
						stringLen = stringLen - lineFeed
						if stringLen > 0:
							responseString = responseString[lineFeed:]
					dev.updateStateOnServer("aux" + auxID, value=auxState)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("AUX" + auxID + ": " + auxState)

				### UNITS command was sent, gets temperature units.
				###	If changed, Jandy will update setpoints, but need to request pool/spa/solar temp updates.
				elif responseString[:11] == "!00 UNITS =":
					currentUnits = dev.states["tempUnits"]
					if responseString[12] == "F":
						tempUnits = u"\u00B0F"
#						tempUnits = "F"		# Can alternatively be used for debugging, to readily allow Terminal to "print" states.
					else:
						tempUnits = u"\u00B0C"
#						tempUnits = "C"		# Can alternatively be used for debugging, to readily allow Terminal to "print" states.
					if tempUnits !=  currentUnits and currentUnits != "":
						if dev.deviceTypeId[19:23] == "Pool":
							if devProps["systemChoice"] == "pool":			
								self.queueSerialCmd("#POOLTMP?")
							else:
								self.queueSerialCmd("#SPATMP?")
						else:
							self.queueSerialCmd("#POOLTMP?")
							self.queueSerialCmd("#SPATMP?")
						if devProps["solHeatCheckbox"] == True:			
							self.queueSerialCmd("#SOLTMP?")		# Not sure what solar temperature is reading?  Need beta tester.
					dev.updateStateOnServer("tempUnits", value=tempUnits)
					if dev.states["loggingChoice"] != "logOff":
						indigo.server.log("Temperature units: " + dev.states["tempUnits"])
					stringLen = stringLen - 15
					if stringLen > 0:
						responseString = responseString[15:]

				### COSMSGS command was sent, turning on or off change of state messages from Jandy.		
				elif responseString[:13] == "!00 COSMSGS =":	
					if responseString[14] == "0":
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("COSMSGS are turned off, therefore the Jandy system will not update the plugin on an ongoing basis.")
					elif responseString[14] == "1":
						if dev.states["loggingChoice"] != "logOff":
							indigo.server.log("COSMSGS are turned on, allowing the Jandy system to update the plugin on an ongoing basis.")
					stringLen = stringLen - 17
					if stringLen > 0:
						responseString = responseString[17:]


###				Response for SALT command is apparently only valid as a query, and is also continuously provided by Jandy communication.
###				Only log changes if logging is set to all, as updates are quite frequent, similar to temperatures.
###				NOTE: Need to also remove "PPM" when reducing the string length.
				elif responseString[:10] == "!00 SALT =":
					lineFeed = string.find(responseString,u"\u000A") + 1
					saltConc = responseString[11:lineFeed-5]
					dev.updateStateOnServer("saltConc", value=saltConc)
					if dev.states["loggingChoice"] == "logAll":
						indigo.server.log("Salt Conc: " + saltConc + " PPM")					
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[19:]

							
###				Responses for MODEL, OPMODE, OPTIONS, VBAT, VERS and S1 commands are valid only as queries.
###				Jandy Model Number is stored as a state, effective plugin version 1.1.
				elif responseString[:11] == "!00 MODEL =":
					jandyModel = responseString[12:16]
					dev.updateStateOnServer("jandyModel", value=jandyModel)
					indigo.server.log("Jandy Model: " + jandyModel)					
					stringLen = stringLen - 18
					if stringLen > 0:
						responseString = responseString[18:]
				elif responseString[:12] == "!00 OPMODE =":	
					lineFeed = string.find(responseString,u"\u000A") + 1
					indigo.server.log("Power Center Operating Mode: " + responseString[13:lineFeed-2])					
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]
				elif responseString[:13] == "!00 OPTIONS =":	
					lineFeed = string.find(responseString,u"\u000A") + 1
					indigo.server.log("Power Center Options DIP Switch: " + responseString[14:lineFeed-2])					
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]
				elif responseString[:10] == "!00 VBAT =":	
					lineFeed = string.find(responseString,u"\u000A") + 1
					batteryValue = float(responseString[11:lineFeed-2])
					if batteryValue < 2048:
						batteryValue = batteryValue/100
						indigo.server.log("Power Center Battery is OK: " + str(batteryValue) + "V")
					else:
						batteryValue = (batteryValue - 2048)/100
						indigo.server.log("Power Center Battery is LOW: " + str(batteryValue) + "V")										
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]
				elif responseString[:10] == "!00 VERS =":	
					indigo.server.log("Jandy Aqualink-RS Version: " + responseString[11:14])					
					stringLen = stringLen - 16
					if stringLen > 0:
						responseString = responseString[16:]
				elif responseString[:8] == "!00 S1 =":	
					lineFeed = string.find(responseString,u"\u000A") + 1
					indigo.server.log("Serial Adaptor DIP Switch 1: " + responseString[9:lineFeed-2])					
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]

###				NOTE: ECHO, RSPFMT, CMDCHR, NRMCHR, ERRCHR can only be executed using the undocumented AQUALINK-RS-CHECK action.
###				Any other use of these commands is prohibited, as they would "break" communication with the Aqualink-RS
				elif responseString[:10] == "!00 ECHO =":
					indigo.server.log(responseString[4:12])					
					stringLen = stringLen - 14
					if stringLen > 0:
						responseString = responseString[14:]
				elif responseString[:12] == "!00 RSPFMT =":	
					indigo.server.log(responseString[4:14])					
					stringLen = stringLen - 16
					if stringLen > 0:
						responseString = responseString[16:]
				elif responseString[:12] == "!00 CMDCHR =":	
					indigo.server.log(responseString[4:15] + ", " + chr(int(responseString[13:15])))					
					stringLen = stringLen - 17
					if stringLen > 0:
						responseString = responseString[17:]
				elif responseString[:12] == "!00 NRMCHR =":	
					indigo.server.log(responseString[4:15] + ", " + chr(int(responseString[13:15])))					
					stringLen = stringLen - 17
					if stringLen > 0:
						responseString = responseString[17:]
				elif responseString[:12] == "!00 ERRCHR =":	
					indigo.server.log(responseString[4:15] + ", " + chr(int(responseString[13:15])))					
					stringLen = stringLen - 17
					if stringLen > 0:
						responseString = responseString[17:]

				### Error response?
				elif responseString[:1] == "?":
					indigo.server.log("Jandy error: " + responseString)
					lineFeed = string.find(responseString,u"\u000A") + 1
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]

				### Undecoded message - will update to include additional "acceptable" responses, if needed.		
				else:
					lineFeed = string.find(responseString,u"\u000A") + 1
					indigo.server.log("Undecoded message from Aqualink: " + responseString[:lineFeed])
					stringLen = stringLen - lineFeed
					if stringLen > 0:
						responseString = responseString[lineFeed:]


	##############################################################################################
	# Convert LED command response into a string set of characters, 
	# excluding "!00 LEDS = " header and carraige return.
	def decodeLED(self, devId, responseString):
		
		########################################				
		lineFeed = string.find(responseString,u"\u000A") + 1
		responseLen = len(responseString) - lineFeed
		stringLED = responseString[11:lineFeed]
		responseLen = len(responseString) - 14
		self.plugin.debugLog(str(lineFeed) + ", " + stringLED + ", " + str(responseLen))

		### Get fresh copy of device states, as they may have changed, and some states
		### are determined / set based on others.
		dev = indigo.devices[devId]
				
		#
		# Convert stringLED characters into LED numeric values for each of 5 Bytes. (See pages 35-37 of Aqualink RS Serial manual)
		#	Each of 5 Bytes encodes 4 "LED," with 2 bits per LED, for a total of 20 LED.
		#	Byte 1 encodes LED 1-4, Byte 2 encodes LED 5-7, Byte 3 encodes LED 8-12, Byte 4 encodes LED 13-16, Byte 5 encodes LED 17-20
		#		LED 1,5,9,13,17 use bits 1-2 (1,2) of their corresponding Bytes
		#		LED 2,6,10,14,18 use bits 3-4 (4,8) of their corresponding Bytes
		#		LED 3,7,11,15,19 use bits 5-6 (16,32) of their corresponding Bytes
		#		LED 4,8,12,16,20 use bits 7-8 (64,128) of their corresponding Bytes
		#		For each LED, both bits = 0 encodes OFF
		#		For each LED, bit n = 1, bit n+1 =0 encodes ON
		#		For each LED, bit n = 0, bit n+1 =1 encodes Flash
		#		For each LED, both bits = 1 encodes Slow Flash
		#
		# Byte 1 encodes LED 1-4.
		#	LED 1 (bits 1,2 = 1,2) encodes AUX7 for Aqualink RS-8, nothing for RS-4, RS-6 or RS-2/6.
		#	LED 2 (bits 3,4 = 4,8) encodes nothing for Aqualink RS-4, RS-6, RS-8 or RS-2/6.
		#	LED 3 (bits 5,6 = 16,32) encodes AUX3 or Waterfall (wfall); parse which when setting states below.
		#	LED 4 (bits 7,8 = 64,128) encodes AUX2 or Low speed pump (pumpLo); parse which when setting states.
		#	With only one LED used, Byte 1 simplifies to 0 = AUX2 Off, 64 = AUX2 On; any other value implies some error.
		if stringLED[1] == " ":
			byte1 = int(stringLED[0])
			nextIndex = 2
		else:
			byte1 = int(stringLED[0:2])
			nextIndex = 3

		bit1 = (byte1 & 1) / 1
		bit2 = (byte1 & 2) / 2
		bit3 = (byte1 & 4) / 4
		bit4 = (byte1 & 8) / 8
		bit5 = (byte1 & 16) / 16
		bit6 = (byte1 & 32) / 32
		bit7 = (byte1 & 64) / 64
		bit8 = (byte1 & 128) / 128

		if bit1 == 0 and bit2 == 0:
			aux7 = "off"
		elif bit1 == 1 and bit2 ==0:
			aux7 = "on"

		if bit5 == 0 and bit6 == 0:
			aux3 = "off"
		elif bit5 == 1 and bit6 ==0:
			aux3 = "on"

		if bit7 == 0 and bit8 == 0:
			aux2 = "off"
		elif bit7 == 1 and bit8 ==0:
			aux2 = "on"

		self.plugin.debugLog("byte1=" + str(byte1) + ", aux7=" + aux7 + ", aux3/wfall=" + aux3 + ", aux2/pumpLo=" + aux2)
		
		#
		# Byte 2 encodes LED 5-8.
		#	LED 5 (bits 1,2 = 1,2) encodes AUX1 or Pool Cleaner, with bit2 indicating cleaner switching; parse which when setting states below.
		#	LED 6 (bits 3,4 = 4,8) encodes Spa (Valve) Status; 0 = Spa Off, 4 = Spa On, 8 = Spa Off after Cooldown, 12 = Pool Cooldown??
		#	LED 7 (bits 5,6 = 16,32) encodes Pump Status; 0 = Pump Off, 16 = Pump On, 32 = Pump Waiting for Delay, 48 =?
		#	LED 8 (bits 7,8 = 64,128) encodes AUX7 for Aqualink RS-6, RS-8 or RS-2/6, nothing for RS-4.
		currentIndex = nextIndex
		nextIndex = nextIndex + 1
		if stringLED[nextIndex] == " ":
			byte2 = int(stringLED[currentIndex])
			nextIndex = nextIndex + 1
		else:
			byte2 = int(stringLED[currentIndex:currentIndex+2])
			nextIndex = nextIndex + 2

		bit1 = (byte2 & 1) / 1
		bit2 = (byte2 & 2) / 2
		bit3 = (byte2 & 4) / 4
		bit4 = (byte2 & 8) / 8
		bit5 = (byte2 & 16) / 16
		bit6 = (byte2 & 32) / 32
		bit7 = (byte2 & 64) / 64
		bit8 = (byte2 & 128) / 128

		if bit1 == 0 and bit2 == 0:
			aux1 = "off"
		elif bit1 == 1 and bit2 == 0:
			aux1 = "on"
		elif bit1 == 0 and bit2 == 1 and "cleanr" in dev.states:
			aux1 = "on"

		### The Pool/Spa Separate system setting is spa pump on/off, Combo systems are valve set to pool/spa or cooldown.		
		### For Pool-Spa Only Systems, bit3 and bit4 should always be 0, but valveState will not be used.
		if dev.deviceTypeId != "jandyAqualink-RS-2-6-SepEquip":
			if bit3 == 0 and bit4 == 0:
				valveState = "pool"
			elif bit3 == 1 and bit4 ==0:
				valveState = "spa"
			### These states have not been seen in practice, but are included because they are referenced in the manual.
			elif bit3 == 0 and bit4 == 1:
				valveState = "spaCoolDown"
			elif bit3 == 1 and bit4 == 1:
				valveState = "poolCoolDown"
		else:
			if bit3 == 0 and bit4 == 0:
				spaPumpState = "off"
			elif bit3 == 1 and bit4 ==0:
				spaPumpState = "on"
			elif bit3 == 0 and bit4 == 1:
				spaPumpState = "spaCoolDown"
			elif bit3 == 1 and bit4 == 1:
				pass


		########   The Pool/Spa Separate system setting is pool pump on/off, Combo systems are pump on/off or waiting for delay.		
		if dev.deviceTypeId != "jandyAqualink-RS-2-6-SepEquip":
			if bit5 == 0 and bit6 == 0:
				pumpState = "off"
			elif bit5 == 1 and bit6 == 0:
				pumpState = "on"
			elif bit5 == 0 and bit6 == 1:
				pumpState = "waitingForDelay"
			elif bit5 == 1 and bit6 == 1:
				pass
		else:
			if bit5 == 0 and bit6 == 0:
				poolPumpState = "off"
			elif bit5 == 1 and bit6 == 0:
				poolPumpState = "on"
			elif bit5 == 0 and bit6 == 1:
				poolPumpState = "waitingForDelay"
			elif bit5 == 1 and bit6 == 1:
				pass

		if bit7 == 0 and bit8 == 0:
			aux5 = "off"
		elif bit7 == 1 and bit8 ==0:
			aux5 = "on"

		if dev.deviceTypeId[19:28] == "PoolOrSpa":
			self.plugin.debugLog("byte2=" + str(byte2) + ", aux1/cleanr=" + aux1 + ", pumpState=" + pumpState + ", aux5=" + aux5)
		elif dev.deviceTypeId[19:24] == "Combo":
			self.plugin.debugLog("byte2=" + str(byte2) + ", aux1/cleanr=" + aux1 + ", valveState=" + valveState + ", pumpState=" + pumpState + ", aux5=" + aux5)
		else:
			self.plugin.debugLog("byte2=" + str(byte2) + ", aux1/cleanr=" + aux1 + ", spaPumpState=" + spaPumpState + ", poolPumpState=" + poolPumpState + ", aux5=" + aux5)

		
		#
		# Byte 3 encodes LED 9-12.
		#	LED 9 (bits 1,2 = 1,2) encodes AUX4 for Aqualink RS-6, RS-8 or RS-2/6, nothing for RS-4.
		#	LED 10 (bits 3,4 = 4,8) encodes nothing for Aqualink RS-4, RS-6, RS-8 or RS-2/6. 
		#	LED 11 (bits 5,6 = 16,32) encodes nothing for Aqualink RS-4, RS-6, RS-8 or RS-2/6.
		#	LED 12 (bits 7,8 = 64,128) encodes AUX6 for Aqualink RS-6, RS-8 or RS-2/6, nothing for RS-4.
		currentIndex = nextIndex
		nextIndex = nextIndex + 1
		if stringLED[nextIndex] == " ":
			byte3 = int(stringLED[currentIndex])
			nextIndex = nextIndex + 1
		else:
			byte3 = int(stringLED[currentIndex:currentIndex+2])
			nextIndex = nextIndex + 2

		bit1 = (byte3 & 1) / 1
		bit2 = (byte3 & 2) / 2
		bit3 = (byte3 & 4) / 4
		bit4 = (byte3 & 8) / 8
		bit5 = (byte3 & 16) / 16
		bit6 = (byte3 & 32) / 32
		bit7 = (byte3 & 64) / 64
		bit8 = (byte3 & 128) / 128

		if bit1 == 0 and bit2 == 0:
			aux4 = "off"
		elif bit1 == 1 and bit2 == 0:
			aux4 = "on"

		if bit7 == 0 and bit8 == 0:
			aux6 = "off"
		elif bit7 == 1 and bit8 == 0:
			aux6 = "on"

		self.plugin.debugLog("byte3=" + str(byte3) + ", aux4=" + aux4 + ", aux6=" + aux6)

			
		#
		# Byte 4 encodes LED 13-16
		#	LED 13-14 encodes nothing for Aqualink RS-4, RS-6, RS-8 or RS-2/6.
		#	LED 15 (bits 5,6 = 16,32) encodes Pool Heater Active using bit 5 off/on (16);  bit 6 (32) is unsed.
		#	LED 16 (bits 7,8 = 64,128) encodes Pool Heater Enabled using bit 7 off/on (64), bit 8 (128) is unused.
		#	Byte 4 simplifies to Pool Heater Status, 0 = Off, 16 = Enabled and active, 64 = Enabled but inactive
		currentIndex = nextIndex
		nextIndex = nextIndex + 1
		if stringLED[nextIndex] == " ":
			byte4 = stringLED[currentIndex]
			nextIndex = nextIndex + 1
		else:
			byte4 = stringLED[currentIndex:currentIndex+2]
			nextIndex = nextIndex + 2
		
		if byte4 == "0":
			poolHeat = "off"
		elif byte4 == "16":
			poolHeat = "on"
		elif byte4 == "64":
			poolHeat = "enabled"

		self.plugin.debugLog("byte4=" + str(byte4) + ", poolHeat=" + poolHeat)
		
		#
		# Byte 5 encodes LED 17-20
		#	LED 17 (bits 1,2 = 1,2) encodes Spa Heater Active using bit 1 off/on (1);  bit 2 (2) is unused.
		#	LED 18 (bits 3,4 = 4,8) encodes Spa Heater Enabled using bit 3 off/on (4), bit 4 (8) is unused.
		#	LED 19 (bits 5,6 = 16,32) encodes Solar Heater Active or auxX using bit 5 off/on (16);  bit 6 (32) is unsed.
		#	LED 20 (bits 7,8 = 64,128) encodes Solar Heater Enabled using bit 7 off/on (64), bit 8 (128) is unused.

		#	Byte 5 simplifies to Spa Heater Status, 0 = Off, 1 = Enabled and active, 4 = Enabled but inactive
		currentIndex = nextIndex
		if nextIndex == responseLen:
			byte5 = int(stringLED[currentIndex])
		else:
			byte5 = int(stringLED[currentIndex:currentIndex+2])

		bit1 = (byte5 & 1) / 1
		bit2 = (byte5 & 2) / 2
		bit3 = (byte5 & 4) / 4
		bit4 = (byte5 & 8) / 8
		bit5 = (byte5 & 16) / 16
		bit6 = (byte5 & 32) / 32
		bit7 = (byte5 & 64) / 64
		bit8 = (byte5 & 128) / 128

		spaHeat = "off"
		if bit1 == 1:
			spaHeat = "on"
		elif bit3 == 1:
			spaHeat = "enabled"

		auxXsolHeat = "off"
		if bit5 == 1:
			auxXsolHeat = "on"
		elif bit7 == 1:
			auxXsolHeat = "enabled"

		self.plugin.debugLog("byte5=" + str(byte5) + ", spaHeat=" + spaHeat + ", auxX/solHeat=" + auxXsolHeat)


		### Update and log state values on server based on LED command values, if changed,
		# and based on device configuration stored in states.
		
		### For each of the states, get current stored state, check vs. state decoded from LED command
		###		and only update on the server if the state has changed.
		###	NOTE: For optional states (AUX, cleanr,etc.), presence in dev.states determines whether it is installed.

		### AUX1 can be used for AUX1 or CLEANR (Pool cleaner) or nothing
		if "aux1" in dev.states:
			currentState = dev.states["aux1"]
			if aux1 != currentState:
				dev.updateStateOnServer("aux1", value=aux1)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("AUX1: " + aux1)
		if "cleanr" in dev.states:
			currentState = dev.states["cleanr"]
			if aux1 != currentState:
				dev.updateStateOnServer("cleanr", value=aux1)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Pool cleaner: " + aux1)

		### AUX2 can be used for AUX2 or PUMPLO (Low speed pump) or nothing
		if "aux2" in dev.states:
			currentState = dev.states["aux2"]
			if aux2 != currentState:
				dev.updateStateOnServer("aux2", value=aux2)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("AUX2: " + aux2)
		if "pumpLo" in dev.states:
			currentState = dev.states["pumpLo"]
			if aux2 != currentState:
				dev.updateStateOnServer("pumpLo", value=aux2)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Low speed pump: " + aux2)

		### AUX3 can be used for AUX3 or WFALL (Waterfall) or nothing
		if "aux3" in dev.states:
			currentState = dev.states["aux3"]
			if aux3 != currentState:
				dev.updateStateOnServer("aux3", value=aux3)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("AUX3: " + aux3)
		if "wfall" in dev.states:
			currentState = dev.states["wfall"]
			if aux3 != currentState:
				dev.updateStateOnServer("wfall", value=aux3)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Waterfall: " + aux3)


		### AUX 4-5 available for RS-6, RS-8 or RS2/6 models
		if dev.deviceTypeId[:18] != "jandyAqualink-RS-4":

			### AUX4 can be used for AUX4 or nothing
			if "aux4" in dev.states:
				currentState = dev.states["aux4"]
				if aux4 != currentState:
					dev.updateStateOnServer("aux4", value=aux4)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("AUX4: " + aux4)
	
			### AUX5 can be used for AUX5 or nothing
			if "aux5" in dev.states:
				currentState = dev.states["aux5"]
				if aux5 != currentState:
					dev.updateStateOnServer("aux5", value=aux5)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("AUX5: " + aux5)
	
		### AUX 6 available only for RS-8 or RS2/6 models
		if dev.deviceTypeId[:18] != "jandyAqualink-RS-4" and dev.deviceTypeId[:18] != "jandyAqualink-RS-6":

			### AUX6 can be used for AUX6 or nothing
			if "aux6" in dev.states:
				currentState = dev.states["aux6"]
				if aux6 != currentState:
					dev.updateStateOnServer("aux6", value=aux6)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("AUX6: " + aux6)

		### AUX 7 available only for RS-8 models
		### AUX7 can be used for AUX7 or nothing
		if dev.deviceTypeId[:18] == "jandyAqualink-RS-8":
			if "aux7" in dev.states:
				currentState = dev.states["aux7"]
				if aux7 != currentState:
					dev.updateStateOnServer("aux7", value=aux7)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("AUX7: " + aux7)

		### AUXX can be used for AUXX or SOLHT (Solar Heat) or nothing
		if "auxX" in dev.states:
			currentState = dev.states["auxX"]
			if auxXsolHeat != currentState:
				dev.updateStateOnServer("auxX", value=auxXsolHeat)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("auxX: " + auxXsolHeat)
		if "solHeat" in dev.states:
			currentState = dev.states["solHeat"]
			if auxXsolHeat != currentState:
				dev.updateStateOnServer("solHeat", value=auxXsolHeat)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Solar Heat: " + auxXsolHeat)

		### Valve state for RS-4, RS-6 and RS-8 combo systems only.
		if dev.deviceTypeId[19:24] == "Combo":
			currentState = dev.states["valveState"]
			if valveState != currentState:
				dev.updateStateOnServer("valveState", value=valveState)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Valves switched/switching to " + valveState)

		### Pump state for RS-4, RS-6 and RS-8. 
		if dev.deviceTypeId != "jandyAqualink-RS-2-6-SepEquip":
			currentState = dev.states["pumpState"]
			if pumpState != currentState:
				dev.updateStateOnServer("pumpState", value=pumpState)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Pump: " + pumpState)

		### Pool pump state for RS-2/6.
		if dev.deviceTypeId == "jandyAqualink-RS-2-6-SepEquip":
			currentState = dev.states["poolPumpState"]
			if poolPumpState != currentState:
				dev.updateStateOnServer("poolPumpState", value=poolPumpState)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Pool pump: " + poolPumpState)

		### Spa pump state for RS-2/6.
		if dev.deviceTypeId == "jandyAqualink-RS-2-6-SepEquip":
			currentState = dev.states["spaPumpState"]
			if spaPumpState != currentState:
				dev.updateStateOnServer("spaPumpState", value=spaPumpState)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Spa pump: " + spaPumpState)

		### Pool / spa heater states. One or the other, not both, for Pool-Spa Only Systems.
		if dev.deviceTypeId[19:28] == "PoolOrSpa":
			devProps = dev.pluginProps
			if devProps["systemChoice"] == "pool":
				currentState = dev.states["poolHeat"]
				if poolHeat != currentState:
					dev.updateStateOnServer("poolHeat", value=poolHeat)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("Pool heat: " + poolHeat)
			else:
				currentState = dev.states["spaHeat"]
				if spaHeat != currentState:
					dev.updateStateOnServer("spaHeat", value=spaHeat)
				if dev.states["loggingChoice"] != "logOff":
					indigo.server.log("Spa heat: " + spaHeat)
		else:
			currentState = dev.states["poolHeat"]
			if poolHeat != currentState:
				dev.updateStateOnServer("poolHeat", value=poolHeat)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Pool heat: " + poolHeat)
			currentState = dev.states["spaHeat"]
			if spaHeat != currentState:
				dev.updateStateOnServer("spaHeat", value=spaHeat)
			if dev.states["loggingChoice"] != "logOff":
				indigo.server.log("Spa heat: " + spaHeat)
